
/* S2 cpu micro architecture simulator

   5 December 2001

   modify from s1 sim

   Computer Architecture 1997, 1998

   microprogram version 		13 Feb 2005

   P. Chongstitvatana
   Department of Computer Engineering
   Chulalongkorn University
*/

#include "s2.h"

int R[32], Ir, Mdr, Mar, Pc, T, Z, S, M[20000];
int runflag, savePc;

void ifetch(void){      	/* instruction fetch */
	Mar = Pc;
	Mdr = M[Mar]; Pc = Pc + 1;
}

void decode(void){
	Ir = Mdr;
}

/* alu operation input : a, b
 affect condition code Z,S
 carry and overflow are not implemented */

int alu(int op, int a, int b){
  int t;
  switch(op) {
  case FADD: t = a + b; break;
  case FSUB: t = a - b; break;
  case FMUL: t = a * b; break;
  case FDIV: t = a / b; break;
  case FAND: t = a & b; break;
  case FOR: t = a | b; break;
  case FXOR: t = a ^ b; break;
  case FSHL: t = a << 1; break;
  case FSHR: t = a >> 1; break;
  case FP1: t = a; break;
  case FP2: t = b; break;
  }
  if( t == 0 ) Z = 1; else Z = 0;
  if( t < 0 )  S = 1; else S = 0;
  return t;
}
// map op to alu functions
int mapop(int op, int xop){
	switch(op){
	case ADDI:	return FADD;
	case SUBI:	return FSUB;
	case MULI:  return FMUL;
	case DIVI:  return FDIV;
	case ANDI:  return FAND;
	case ORI:	return FOR;
	case XORI:	return FXOR;
	case XOP:	return xop;
	defaut: error("undefine alu op");
	}
	return 0;
}
// effective address adder, do not affect flag
int adder(int a, int b){
	return a + b;
}

int testCc(int cond ){	/* test condition code */
  switch( cond ){
  case ALWAYS: return 1;
  case EQ: return Z;
  case NEQ: return Z == 0;
  case LT: return S && (Z == 0);
  case LE: return S || Z;
  case GE: return (S == 0) || Z;
  case GT: return (S == 0) && (Z == 0);
  default: error("undefine conditional code");
  }
  return 0;
}
/* write register, check for R[0] */
void wR(int rn, int value){
	if(rn != 0) R[rn] = value;
}

void opRR(char op){
	T = alu(op,R[IRr2()],R[IRr3()]);
	wR(IRr1(),T);
}

void opRI(char op){
	T = alu(op,R[IRr2()],IRdisp());
	wR(IRr1(),T);
}

void countclock(void);
void xop(void);
void show(void);

// hardwired control unit
// execute one instruction
void run(void){
	savePc = Pc;
    ifetch();
	decode();
    countclock();
    switch( IRop() ) {
    case LDA:
		Mar = IRads();
		Mdr = M[Mar];
	    wR(IRr1(),Mdr);
	    break;
	case LDD:
//		T = alu('+',R[IRr2()],IRdisp());
//		Mar = T;
		Mar = adder(R[IRr2()],IRdisp());  // not affect flag
		Mdr = M[Mar];
		wR(IRr1(),Mdr);
		break;
	case LDI:
		wR(IRr1(),IRdisp());
		break;
	case STA:
		Mar = IRads();
		Mdr = R[IRr1()];
		M[Mar] = Mdr;
	    break;
	case STD:
//		T = alu('+',R[IRr2()],IRdisp());
//		Mar = T;
		Mar = adder(R[IRr2()],IRdisp());
		Mdr = R[IRr1()];
		M[Mar] = Mdr;
		break;
    case JMP:
		if( testCc(IRr1()) )
			Pc = IRads();
	    break;
	case JAL:
		wR(IRr1(),Pc);
		Pc = IRads();
		break;
	case ADDI: opRI(FADD); break;
	case SUBI: opRI(FSUB); break;
	case MULI: opRI(FMUL); break;
	case DIVI: opRI(FDIV); break;
	case ANDI: opRI(FAND); break;
	case ORI:  opRI(FOR); break;
	case XORI: opRI(FXOR); break;
	case XOP: xop(); break;   /* extended op */
    default: error("undefine op");
	}
	if(disp) show();
}

void xop(void){
	switch( IRxop() ) {
	case ADD: opRR(FADD); break;
	case SUB: opRR(FSUB); break;
	case MUL: opRR(FMUL); break;
	case DIV: opRR(FDIV); break;
	case AND: opRR(FAND); break;
	case OR:  opRR(FOR); break;
	case XOR: opRR(FXOR); break;
	case SHL: opRR(FSHL); break;
	case SHR: opRR(FSHR); break;
	case LDX:
//		T = alu('+',R[IRr2()],R[IRr3()]);
//		Mar = T;
		Mar = adder(R[IRr2()],R[IRr3()]); // not affect flag
		Mdr = M[Mar];
		wR(IRr1(),Mdr);
		break;
	case STX:
//		T = alu('+',R[IRr2()],R[IRr3()]);
//		Mar = T;
		Mar = adder(R[IRr2()],R[IRr3()]);
		Mdr = R[IRr1()];
		M[Mar] = Mdr;
		break;
	case JR: Pc = R[IRr1()]; break;
	case TRAP: trap(IRr1(),IRr2()); break;
	default: error("undefine op");
	}
}

int main(int argc, char *argv[]){
  if( argc < 2 ) {
    printf("usage : s2m objfile\n");
    exit(0);
  }
  loadprogram(argv[1]);
  clock = 0;
  ninst = 0;
  Pc = 0;
  Z = 0;
  S = 0;
  R[0] = 0;
  runflag = 1;
//  interp(1);
  readmspec();
  readmpgm();
//  printout();
  mpc = 0;
  initrom();
  interp(2);
  return 0;
}

